/****************************************************************************
Copyright (c) 2010-2012 cocos2d-x.org
Copyright (c) 2013-2014 Chukong Technologies Inc.

http://www.cocos2d-x.org

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
****************************************************************************/

#include "base/CCPlatformConfig.h"
#if CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID

#include "CCFileUtilsAndroid.h"
#include "platform/CCCommon.h"
#include "jni/Java_org_cocos2dx_lib_Cocos2dxHelper.h"
#include "android/asset_manager.h"
#include "android/asset_manager_jni.h"

#include <stdlib.h>

#define  LOG_TAG    "CCFileUtilsAndroid.cpp"
#define  LOGD(...)  __android_log_print(ANDROID_LOG_DEBUG,LOG_TAG,__VA_ARGS__)



#include "base/ZipUtils.h"


using namespace std;

NS_CC_BEGIN

// record the zip on the resource path for the obb
static ZipFile *s_pZipFile = NULL;

AAssetManager* FileUtilsAndroid::assetmanager = nullptr;

void FileUtilsAndroid::setassetmanager(AAssetManager* a) {
    if (nullptr == a) {
        LOGD("setassetmanager : received unexpected nullptr parameter");
        return;
    }

    cocos2d::FileUtilsAndroid::assetmanager = a;
}

FileUtils* FileUtils::getInstance()
{
    if (s_sharedFileUtils == nullptr)
    {
        s_sharedFileUtils = new FileUtilsAndroid();
        if(!s_sharedFileUtils->init())
        {
          delete s_sharedFileUtils;
          s_sharedFileUtils = nullptr;
          CCLOG("ERROR: Could not init CCFileUtilsAndroid");
        }
        
        std::string resourcePath = getApkPath();
        s_pZipFile = new ZipFile(resourcePath, "assets/");
    }
    return s_sharedFileUtils;
}

FileUtilsAndroid::FileUtilsAndroid()
{
}

FileUtilsAndroid::~FileUtilsAndroid()
{
    CC_SAFE_DELETE(s_pZipFile);
}

bool FileUtilsAndroid::init()
{
    _defaultResRootPath = "assets/";
    return FileUtils::init();
}

bool FileUtilsAndroid::isFileExistInternal(const std::string& strFilePath) const
{
    if (strFilePath.empty())
    {
        return false;
    }

    bool bFound = false;
    
    // Check whether file exists in apk.
    if (strFilePath[0] != '/')
    {
        //CCLOG("isFileExistInternal - path start with / so should be in apk or obb");
        const char* s = strFilePath.c_str();
        
        // Found "assets/" at the beginning of the path and we don't want it
        if (strFilePath.find(_defaultResRootPath) == 0) s += strlen("assets/");
        
        if (FileUtilsAndroid::assetmanager) {
            AAsset* aa = AAssetManager_open(FileUtilsAndroid::assetmanager, s, AASSET_MODE_UNKNOWN);
            if (aa)
            {
                //CCLOG("[AssetManager] ... in APK %s, found = true!", strFilePath.c_str());
                bFound = true;
                AAsset_close(aa);
            }
            else
            {
                //CCLOG("[AssetManager] ... in APK %s, found = false!", strFilePath.c_str());
           
        
                std::string strPath = strFilePath;
                if (strPath.find(_defaultResRootPath) != 0)
                {// Didn't find "assets/" at the beginning of the path, adding it.
                    strPath.insert(0, _defaultResRootPath);
                }
        
                if (s_pZipFile->fileExists(strPath))
                {
                    //CCLOG("[s_pZipFile] ... in OBB %s, found = true!", strFilePath.c_str());
                    bFound = true;
                }
                else {
                    //CCLOG("[s_pZipFile] ... in OBB %s, found = false!", strFilePath.c_str());
                }
	    
	     }
        }
    }
    else
    {
        //CCLOG("isFileExistInternal - path does not start with / so should be in file system");
        FILE *fp = fopen(strFilePath.c_str(), "r");
        if(fp)
        {
            bFound = true;
            fclose(fp);
        }
    }
    return bFound;
}

bool FileUtilsAndroid::isAbsolutePath(const std::string& strPath) const
{
    // On Android, there are two situations for full path.
    // 1) Files in APK, e.g. assets/path/path/file.png
    // 2) Files not in APK, e.g. /data/data/org.cocos2dx.hellocpp/cache/path/path/file.png, or /sdcard/path/path/file.png.
    // So these two situations need to be checked on Android.
    if (strPath[0] == '/' || strPath.find(_defaultResRootPath) == 0)
    {
        return true;
    }
    return false;
}

Data FileUtilsAndroid::getData(const std::string& filename, bool forString)
{
    //LOGD("FileUtilsAndroid::getData");
    if (filename.empty())
    {
        return Data::Null;
    }
    
    unsigned char* data = nullptr;
    ssize_t size = 0;
    
    string fullPath = fullPathForFilename(filename);    
    
    if (fullPath[0] != '/')
    {
        //CCLOG("getData - path does not start with / so should be in file system");
        string relativePath = string();
        
        size_t position = fullPath.find("assets/");
        if (0 == position) {
            // "assets/" is at the beginning of the path and we don't want it
            relativePath += fullPath.substr(strlen("assets/"));
        } else {
            relativePath += fullPath;
        }
        //LOGD("relative path = %s", relativePath.c_str());

        if (nullptr == FileUtilsAndroid::assetmanager) {
            //LOGD("... FileUtilsAndroid::assetmanager is nullptr");
            return Data::Null;
        }

        
        // read asset data
        AAsset* asset =
            AAssetManager_open(FileUtilsAndroid::assetmanager,
                               relativePath.c_str(),
                               AASSET_MODE_UNKNOWN);
        if (nullptr == asset) {
            //LOGD("asset is nullptr so not in apk");
            //CCLOG("GETTING FILE RELATIVE DATA: %s", fullPath.c_str());
            Data ret;
            unsigned char* buffer = nullptr;
            
            buffer  = this->getFileData(fullPath.c_str(),"r",&size);
            ret.fastSet(buffer, size);   
            if(ret.isNull()) {
                 CCLOG("Data in assets failed for %s",filename.c_str());
                  return Data::Null;                 
            }
            else {
                return ret;
            }
            
            return Data::Null;
        }

        off_t fileSize = AAsset_getLength(asset);

        if (forString)
        {
            data = (unsigned char*) malloc(fileSize + 1);
            data[fileSize] = '\0';
        }
        else
        {
            data = (unsigned char*) malloc(fileSize);
        }

        int bytesread = AAsset_read(asset, (void*)data, fileSize);
        size = bytesread;

        AAsset_close(asset);
    }
    else
    {
        //CCLOG("getData - path does not start with / so should be in file system"); 
        do
        {
            // read rrom other path than user set it
            //CCLOG("GETTING FILE ABSOLUTE DATA: %s", filename);
            const char* mode = nullptr;
            if (forString)
                mode = "rt";
            else
                mode = "rb";

            FILE *fp = fopen(fullPath.c_str(), mode);
            CC_BREAK_IF(!fp);
            
            long fileSize;
            fseek(fp,0,SEEK_END);
            fileSize = ftell(fp);
            fseek(fp,0,SEEK_SET);
            if (forString)
            {
                data = (unsigned char*) malloc(fileSize + 1);
                data[fileSize] = '\0';
            }
            else
            {
                data = (unsigned char*) malloc(fileSize);
            }
            fileSize = fread(data,sizeof(unsigned char), fileSize,fp);
            fclose(fp);
            
            size = fileSize;
        } while (0);
    }
    
    Data ret;
    if (data == nullptr || size == 0)
    {
        std::string msg = "Get data from file(";
        msg.append(filename).append(") failed!");
        CCLOG("%s", msg.c_str());
    }
    else
    {
        ret.fastSet(data, size);
    }

    return ret;
}

std::string FileUtilsAndroid::getStringFromFile(const std::string& filename)
{
    Data data = getData(filename, true);
    if (data.isNull())
        return "";

    std::string ret((const char*)data.getBytes());
    return ret;
}
    
Data FileUtilsAndroid::getDataFromFile(const std::string& filename)
{
    return getData(filename, false);
}

unsigned char* FileUtilsAndroid::getFileData(const std::string& filename, const char* mode, ssize_t * size)
{    
    //LOGD("FileUtilsAndroid::getFileData");
    unsigned char * pData = 0;

    if ( filename.empty() || (! mode) )
    {
        return 0;
    }
    
    string fullPath = fullPathForFilename(filename);
   
    if (fullPath[0] != '/')
    {
        //CCLOG("GETTING FILE RELATIVE DATA: %s", fullPath.c_str());
        pData = s_pZipFile->getFileData(fullPath.c_str(), size);
    }
    else
    {
        do 
        {
            // read rrom other path than user set it
	    //CCLOG("GETTING FILE ABSOLUTE DATA: %s", filename.c_str());
            FILE *fp = fopen(filename.c_str(), mode);            
            CC_BREAK_IF(!fp);

            unsigned long sizeL;
            fseek(fp,0,SEEK_END);
            sizeL = ftell(fp);            
            fseek(fp,0,SEEK_SET);            
            pData = new unsigned char[sizeL];
            size_t sizeT = fread(pData,sizeof(unsigned char), sizeL,fp);
            fclose(fp);
            
            if (sizeT)
            {
                *size = sizeL;
            }            
        } while (0);        
    }
    if (! pData)
    {
        std::string msg = "Get data from file(";
        msg.append(filename.c_str()).append(") failed!");
        CCLOG(msg.c_str());
    }

    return pData;
}

string FileUtilsAndroid::getWritablePath() const
{
    // Fix for Nexus 10 (Android 4.2 multi-user environment)
    // the path is retrieved through Java Context.getCacheDir() method
    string dir("");
    string tmp = getFileDirectoryJNI();

    if (tmp.length() > 0)
    {
        dir.append(tmp).append("/");

        return dir;
    }
    else
    {
        return "";
    }
}

NS_CC_END

#endif // CC_TARGET_PLATFORM == CC_PLATFORM_ANDROID
